import pytest
import numpy as np
from NonlinearTMM import SecondOrderNLTMM, Material

@pytest.fixture()
def tmm():
    wlP1 = 1000e-9
    wlP2 = 1000e-9
    polP1 = "s"
    polP2 = "s"
    polGen = "s"
    Ly = 2e-3
    w0P1 = 4000e-6
    w0P2 = 4000e-6
    pwrP1 = 1.1
    E0P2 = 12.0
    
    crystalD = 50e-6
    
    # Define materials
    wlsCrystal = np.array([400e-9, 1100e-9])
    nsCrystal = np.array([1.54, 1.53], dtype = complex)
    prism = Material.Static(1.5)
    crystal = Material(wlsCrystal, nsCrystal)
    dielectric = Material.Static(1.6)
    crystal.chi2.Update(d11 = 1e-12, d22 = 1e-12, d33 = 1e-12)
    
    # Init SecondOrderNLTMM
    tmm = SecondOrderNLTMM()
    tmm.P1.SetParams(wl = wlP1, pol = polP1, beta = 0.0, I0 = pwrP1 / (Ly * w0P1))
    tmm.P2.SetParams(wl = wlP2, pol = polP2, beta = 0.0, overrideE0 = True, E0 = E0P2)
    tmm.Gen.SetParams(pol = polGen)
    
    # Add layers
    tmm.AddLayer(float("inf"), prism)
    tmm.AddLayer(crystalD, crystal)
    tmm.AddLayer(float("inf"), dielectric)

    # Init waves
    waveP1Params = {"waveType": "tukey", "pwr": pwrP1, "w0": w0P1, "Ly": Ly, "dynamicMaxXCoef": 5, "nPointsInteg": 200, "maxPhi": np.radians(10.0)}
    waveP2Params = {"waveType": "planewave", "w0": w0P2, "Ly": Ly, "overrideE0": True, "E0": E0P2}
    tmm.P1.wave.SetParams(**waveP1Params)
    tmm.P2.wave.SetParams(**waveP2Params)
    
    return tmm

@pytest.fixture(params = ["tukey", "gaussian", "planewave"])
def waveType(request):
    return request.param

@pytest.fixture(params = ["p", "s"])
def polGen(request):
    return request.param

@pytest.fixture(params = ["p", "s"])
def polP1(request):
    return request.param

@pytest.fixture(params = ["p", "s"])
def polP2(request):
    return request.param

def test_SecondOrderNLTMMWaveSweep(tmm, waveType, polP1, polP2, polGen):
    betas = np.linspace(0.0, 0.99, 10)
    
    tmm.P1.pol = polP1
    tmm.P2.pol = polP2
    tmm.Gen.pol = polGen
    tmm.P1.wave.waveType = waveType
    
    sr = tmm.Sweep("beta", betas, betas, outP1 = False, outP2 = False)
    srW = tmm.WaveSweep("beta", betas, betas, outP1 = False, outP2 = False)

    np.testing.assert_allclose(srW.Gen.Pr, sr.Gen.Ir * srW.Gen.beamArea, rtol = 1e-2)
    np.testing.assert_allclose(srW.Gen.Pt, sr.Gen.It * srW.Gen.beamArea, rtol = 6e-2)
    
    


from libcpp cimport bool
from eigency.core cimport *
from libcpp.pair cimport pair

#===============================================================================
# Common.h
#===============================================================================
        
cdef extern from "Common.h" namespace "TMM":
    cdef enum PolarizationCpp "TMM::Polarization":
        P_POL,
        S_POL
        
    cdef enum WaveDirectionCpp "TMM::WaveDirection":
        F,
        B,
        TOT
        
    cdef enum NonlinearTmmModeCpp "TMM::NonlinearTmmMode":
        MODE_INCIDENT,
        MODE_NONLINEAR
        
    cdef enum TMMParamCpp "TMM::TMMParam":
        PARAM_WL,
        PARAM_BETA,
        PARAM_POL,
        PARAM_I0,
        PARAM_OVERRIDE_E0,
        PARAM_E0,
        PARAM_MODE,
        PARAM_WAVE_W0,
        PARAM_LAYER_D,
        
    cdef enum NonlinearProcessCpp "TMM::NonlinearProcess":
        SFG,
        DFG,
        SPDC,
        
    cdef enum SweepOutputCpp "TMM::SweepOutput":
        SWEEP_PWRFLOWS,
        SWEEP_ABS,
        SWEEP_ENH,
        SWEEP_I,
        SWEEP_R,
        SWEEP_T,
        SWEEP_P1,
        SWEEP_P2,
        SWEEP_GEN,
        SWEEP_ALL_WAVES,
        SWEEP_ALL_WAVE_PWRS,

#===============================================================================
# Material.h
#===============================================================================

cdef extern from "Material.h" namespace "TMM":

    cdef cppclass Chi2TensorCpp "TMM::Chi2Tensor":
        void SetDistinctFields(bool distinctFields_) except +
        void SetRotation(double phiX_, double phiY_, double phiZ_) except +
        void SetChi2(int i1, int i2, int i3, double value) except +
        void SetD(int i1, int i2, double value) except +
        void Clear() except +
        double GetChi2Element(int i1, int i2, int i3) except +
    
    cdef cppclass MaterialCpp "TMM::Material":
        Chi2TensorCpp chi2
        MaterialCpp(Map[ArrayXd] & wlsExp, Map[ArrayXcd] & nsExp) except +
        double complex GetN(double wl) except +
        bool IsNonlinear() except +

#===============================================================================
# Waves.h
#===============================================================================

cdef extern from "Waves.h" namespace "TMM":
    cdef enum WaveTypeCpp "TMM::WaveType":
        PLANEWAVE,
        GAUSSIANWAVE,
        TUKEYWAVE,
        SPDCWAVE

    cdef cppclass WaveCpp "TMM::Wave":
        void SetWaveType(WaveTypeCpp waveType_) except +
        void SetPwr(double pwr_) except +
        void SetOverrideE0(bool overrideE0_) except +
        void SetE0(double E0_) except +
        void SetW0(double w0_) except +
        void SetMaterial(MaterialCpp *material_) except +
        void SetLy(double Ly_) except +
        void SetA(double a_) except +
        void SetNPointsInteg(int nPointsInteg_) except +
        void SetMaxX(double maxX_) except +
        void EnableDynamicMaxX(bool dynamicMaxX_) except +
        void SetDynamicMaxXCoef(double dynamicMaxXCoef_) except +
        void SetDynamicMaxXAddition(double) except +
        void SetMaxPhi(double maxPhi_) except +
        
        double GetPwr() except +
        bool GetOverrideE0() except +
        double GetE0() except +
        double GetW0() except +
        double GetLy() except +
        double GetA() except +
        int GetNPointsInteg() except +
        double GetMaxX() except +
        bool IsDynamicMaxXEnabled() except +
        double GetDynamicMaxXCoef() except +
        double GetDynamicMaxXAddition() except +
        double GetMaxPhi() except +
        pair[double, double] GetXRange() except +
        ArrayXd GetBetas() except +
        ArrayXd GetPhis() except +
        ArrayXd GetKxs() except +
        ArrayXd GetKzs() except +
        ArrayXd GetFieldProfileXs() except +
        ArrayXd GetFieldProfile() except +
        double GetBeamArea() except +
        ArrayXcd GetExpansionCoefsKx() except +
        
#===============================================================================
# NonlinearLayer.h
#===============================================================================

cdef extern from "NonlinearLayer.h" namespace "TMM":

    cdef cppclass HomogeneousWaveCpp "TMM::HomogeneousWave":
        double complex GetKzF() except +
        double GetKx() except +
        Array2cd GetMainFields(double z) except +

    cdef cppclass NonlinearLayerCpp "TMM::NonlinearLayer":
        HomogeneousWaveCpp* GetHw() except +
        void SetThickness(double d_) except +
        double GetThickness() except +
        double GetIntensity(double z) except +
        double GetAbsorbedIntensity() except +
        double GetSrcIntensity() except +
        
#===============================================================================
# NonlinearTMM.h
#===============================================================================

cdef extern from "NonlinearTMM.h" namespace "TMM":

    cdef cppclass IntensitiesCpp "TMM::Intensities":
        double complex inc, r, t
        double I, R, T
        
    cdef cppclass SweepResultNonlinearTMMCpp "TMM::SweepResultNonlinearTMM":
        ArrayXcd inc, r, t;
        ArrayXd Ii, Ir, It, Ia, enh;
        
        int GetOutmask();

    cdef cppclass WaveSweepResultNonlinearTMMCpp "TMM::WaveSweepResultNonlinearTMM":
        ArrayXd Pi, Pr, Pt, enh, beamArea;
        
        int GetOutmask();

    cdef cppclass FieldsZCpp "TMM::FieldsZ":
        MatrixXcd E, H;

    cdef cppclass FieldsZXCpp "TMM::FieldsZX":
        MatrixXcd Ex, Ey, Ez, Hx, Hy, Hz;
        PolarizationCpp GetPol()

    cdef cppclass NonlinearTMMCpp "TMM::NonlinearTMM":
        NonlinearTMMCpp() except +
        void AddLayer(double d_, MaterialCpp *material_) except +
        NonlinearLayerCpp* GetLayer(int layerNr)  except +
        int LayersCount() except +
       
        void SetWl(double wl_) except +
        void SetBeta(double beta_) except +
        void SetPolarization(PolarizationCpp pol_) except +
        void SetI0(double I0_) except +
        void SetOverrideE0(bool overrideE0_) except +
        void SetE0(double complex E0_) except +
        void SetMode(NonlinearTmmModeCpp mode_) except +
        void SetParam(TMMParamCpp param, double value) except +
        void SetParam(TMMParamCpp param, double complex value) except +
        
        double GetWl() except +
        double GetBeta() except +
        PolarizationCpp GetPolarization() except +
        double GetI0() except +
        bool GetOverrideE0() except +
        double complex GetE0() except +
        NonlinearTmmModeCpp GetMode() except +
        double GetDouble(TMMParamCpp param) except +
        complex GetComplex(TMMParamCpp param) except +
       
        # Plane waves
        void Solve() except +
        IntensitiesCpp GetIntensities() except +
        SweepResultNonlinearTMMCpp* Sweep(TMMParamCpp param, Map[ArrayXd] &, int outmask, int paramLayer, int layerNr, double layerZ) except +
        FieldsZCpp* GetFields(Map[ArrayXd] &, WaveDirectionCpp dir) except +
        FieldsZXCpp* GetFields2D(Map[ArrayXd] &, Map[ArrayXd] &, WaveDirectionCpp dir) except +
        double GetAbsorbedIntensity() except +
        double GetEnhancement(int layerNr, double z) except +
        
        # Waves
        WaveCpp* GetWave() except +
        pair[double, double] WaveGetPowerFlows(int layerNr, double x0, double x1, double z) except +
        double WaveGetEnhancement(int layerNr, double z) except +
        WaveSweepResultNonlinearTMMCpp* WaveSweep(TMMParamCpp param, Map[ArrayXd] &, int outmask, int paramLayer, int layerNr, double layerZ) except +
        FieldsZXCpp* WaveGetFields2D(Map[ArrayXd] &, Map[ArrayXd] &, WaveDirectionCpp) except +
      
        
#===============================================================================
# SecondOrderNLTMM.h
#===============================================================================

cdef extern from "SecondOrderNLTMM.h" namespace "TMM":

    cdef cppclass SecondOrderNLIntensitiesCpp "TMM::SecondOrderNLIntensities":
        IntensitiesCpp P1, P2, Gen;
        
    cdef cppclass SweepResultSecondOrderNLTMMCpp "TMM::SweepResultSecondOrderNLTMM":
        SweepResultNonlinearTMMCpp P1, P2, Gen;
        ArrayXd wlsGen, betasGen;
    
    cdef cppclass WaveSweepResultSecondOrderNLTMMCpp "TMM::WaveSweepResultSecondOrderNLTMM":
        WaveSweepResultNonlinearTMMCpp P1, P2, Gen;
        ArrayXd wlsGen, betasGen;
        
    cdef cppclass SecondOrderNLTMMCpp "TMM::SecondOrderNLTMM":
        SecondOrderNLTMM() except +
        void SetProcess(NonlinearProcessCpp process_) except +
        void SetDeltaWlSpdc(double value) except +
        void SetSolidAngleSpdc(double value) except +
        void SetDeltaThetaSpdc(double value) except +
        void AddLayer(double d_, MaterialCpp *material_) except +
        NonlinearTMMCpp* GetP1() except +
        NonlinearTMMCpp* GetP2() except +
        NonlinearTMMCpp* GetGen() except +
        double GetDeltaWlSpdc() except +
        double GetSolidAngleSpdc() except +
        double GetDeltaThetaSpdc() except +
        void UpdateGenParams() except +
        
        void Solve() except +
        SecondOrderNLIntensitiesCpp GetIntensities() except +
        SweepResultSecondOrderNLTMMCpp* Sweep(TMMParamCpp param, Map[ArrayXd] &, Map[ArrayXd] &, int outmask, int paramLayer, int layerNr, double layerZ) except +
        
        pair[double, double] WaveGetPowerFlows(int layerNr, double x0, double x1, double z) except +
        WaveSweepResultSecondOrderNLTMMCpp * WaveSweep(TMMParamCpp param, Map[ArrayXd] & valuesP1, Map[ArrayXd] & valuesP2, int outmask, int paramLayer, int layerNr, double layerZ) except +
        FieldsZXCpp * WaveGetFields2D(Map[ArrayXd]& zs, Map[ArrayXd]& xs, WaveDirectionCpp dir) except +



import numpy as np
import pytest
import LabPy
from NonlinearTMM import Material

@pytest.fixture(scope = "module", params = ["main/Ag/Johnson", "main/Au/Johnson"])
def materials(request):    
    labpyMaterial = LabPy.Material(request.param)
    tmmMaterial = Material.FromLabPy(labpyMaterial) 
    return (tmmMaterial, labpyMaterial)

@pytest.fixture(scope = "module")
def chi2s():
    dValues = {"d11": 1e-12, "d22": 2e-12, "d33": 3e-12, "d12": 4e-12, "d23": 5e-12}
    materialCpp = Material.FromLabPy(LabPy.Material("Static", n = 1.0))
    chi2Cpp = materialCpp.chi2
    chi2Cpp.Update(**dValues)
    chi2Py = LabPy.Chi2Tensor(**dValues)
    return chi2Cpp, chi2Py

@pytest.fixture(params = np.linspace(300e-9, 1500e-9, 100))
def wl(request):
    return request.param

@pytest.fixture(params = [0.0, 0.5, 1.2])
def phiX(request):
    return request.param

@pytest.fixture(params = [0.0, 0.5, 1.2])
def phiY(request):
    return request.param

@pytest.fixture(params = [0.0, 0.5, 1.2])
def phiZ(request):
    return request.param

def test_RefractiveIndex(materials, wl):
    tmmMaterial, labpyMaterial = materials
    np.testing.assert_almost_equal(tmmMaterial.GetN(wl), labpyMaterial(wl))
    
def test_Chi2Tensor(chi2s, phiX, phiY, phiZ):
    chi2Cpp, chi2Py = chi2s
    chi2Cpp.Update(phiX = phiX, phiY = phiY, phiZ = phiZ)
    chi2Py.Update(phiX = phiX, phiY = phiY, phiZ = phiZ)
    
    tensorPy = chi2Py.GetChi2().real
    tensorCpp = chi2Cpp.GetChi2Tensor()
    np.testing.assert_allclose(tensorPy, tensorCpp)
    
    


import numpy as np
from os import path
from scipy.interpolate import interp1d
import fnmatch
import yaml
import bisect
import os
from LabPy import Core  # @UnresolvedImport

materialsDir = path.join(path.dirname(path.realpath(__file__)), r"../materials") 
if not path.isdir(materialsDir):
    materialsDir = os.environ["PYLAB_MATERIALS_DIR"]

#===============================================================================
# Methods
#===============================================================================

def literal_unicode_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')

def GetMaterialsList():
    """Generates list of all materials in database.
    
    Returns:
        list: list of paths to all available materials
    
    """
    res = ["Static"]
    for root, _ ,files in os.walk(materialsDir):
        root = root[len(materialsDir)+1:]
        files[:] = fnmatch.filter(files,'*.yml')
        files = [path.join(root, fn)[:-4] for fn in files]
        res.extend(files)
    return res

def SaveMaterialFile(filename, wls, n):
    stream = open(filename, "w")
    
    data = {}
    data["BOOK"] = "a"
    data["BOOK LONG"] = "b"
    data["PAGE"] = "c"
    data["PAGE LONG"] = "d"
    data["REFERENCES"] = "e"
    data["COMMENTS"] = "f"
    
    nkData = []
    for wl, n in zip(wls, n):
        nkData.append("%s %s %s" % (1e6 * wl, n.real, n.imag))
    nkData = "\n".join(nkData)
    
    class literal_unicode(str): pass

    yaml.add_representer(literal_unicode, literal_unicode_representer)
    data["DATA"] = {"type": "nk", "data": literal_unicode(nkData)}
    
    yaml.dump(data, stream, default_flow_style = False)
    stream.close()

def MaterialFromConf(conf):
    param, kwargParams = conf
    res = Material(param, **kwargParams)
    return res

#===============================================================================
# Classes
#===============================================================================

class Material(Core.ParamsBaseClass):
    """This class reads and processes the refractiveindex.info database.
    
    Args:
        materialFile (str): material filename or Static
        
    Attributes:
        name (str): The name of material
        longName (str): The long name of material
        page (str): The page of material
        longPage (str): The long name of page of material
    
    """
   
    def __init__(self, materialFile, boundsError = True, **kwargs):
        self._params = ["kAdditional"]
        self.materialFile = materialFile
        self.wlExp = None
        self.nExp = None
        self.kExp = None
        self.kAdditional = 0.0
        self.boundsError = boundsError
        self.isFormula = False
        if materialFile == "Static":
            self.n = 1.0
            self.k = 0.0
            self.name = "Static"
            self.nFunc = lambda wl: np.ones_like(wl) * self.n
            self.kFunc = lambda wl: np.ones_like(wl) * self.k
            self._params += ["n", "k"]
        else:
            self._LoadFromFile(materialFile)

        super(Material, self).__init__(**kwargs)
        
    def _LoadFromFile(self, materialFile):
        self.filename = path.join(materialsDir, "%s.yml" % (materialFile))
        
        stream = open(self.filename, "r")
        self.rawData = yaml.load(stream)

        self.nFunc = None
        self.kFunc = lambda wl: 0.0
        
        self.name = self.rawData.get("BOOK", None)
        self.longName = self.rawData.get("BOOK LONG", None)
        self.page = self.rawData.get("PAGE", None)
        self.longPage = self.rawData.get("PAGE LONG", None)
        self.comments = self.rawData.get("COMMENTS", None)
        self.references = self.rawData.get("REFERENCES", None)

        self._ReadDatapoints()
        self._ReadFormulas()

        
    def __call__(self, wl):
        """Returns (interpolates if needed) complex refractive index at wavelength wl.
        
        Args:
            wl (float): wavelength of light
        
        Returns:
            complex: complex refractive index
        
        """
        res = self.nFunc(wl) + 1.0j * self.kFunc(wl) + 1.0j * self.kAdditional
        return res
  
    def GetN(self, wlFr = -float("inf"), wlTo = -float("inf")):
        """Returns complex refractive index datapoints in range wlFr - wlTo.

        Args:
           wlFr (float, optional):  The minimum wavelength.
           wlTo (float, optional):  The maximum wavelength.

        Returns:
           tuple_of_arrays:  wavelength and complex refractive index arrays
        
        """
        index1 = bisect.bisect_left(self.wlExp, wlFr)
        index2 = bisect.bisect_right(self.wlExp, wlTo)

        wls = self.wlExp[index1:index2]
        n = self.nExp[index1:index2] + 1.0j * self.kExp[index1:index2]
        return wls, n

    # Private methods

    def _ReadDatapoints(self):
        if not "DATA" in self.rawData or not "data" in self.rawData["DATA"]:
            return

        dataType = self.rawData["DATA"]["type"]
        dataStr = self.rawData["DATA"]["data"].strip().split("\n")
        
        if dataType == "nk":
            data = np.zeros((len(dataStr), 3))
        elif dataType == "n" or dataType == "k":
            data = np.zeros((len(dataStr), 2))
        else:
            raise ValueError("Unknown data type.")
        
        for i in np.arange(len(dataStr)):
            data[i, :] = np.array(list(map(float, dataStr[i].split())))
        
        if dataType == "nk":
            self.wlExp, self.nExp, self.kExp = data.T
        elif dataType == "n":
            self.wlExp, self.nExp = data.T
        elif dataType == "k":
            self.wlExp, self.kExp = data.T
        else:
            raise ValueError("Unknown data type.")
        
        # Convert wavelength to SI units
        self.wlExp *= 1e-6
        
        # Define interpolation functions
        if not self.nExp is None:
            self.nFunc = interp1d(self.wlExp, self.nExp, bounds_error = self.boundsError)
            
        if not self.kExp is None:
            self.kFunc = interp1d(self.wlExp, self.kExp, bounds_error = self.boundsError)
        
    def _ReadFormulas(self):
        if "FORMULA" in self.rawData:
            dataFormula = self.rawData["FORMULA"]
            wlRange = 1e-6 * np.array(dataFormula["range"].split()).astype(float)
            dispType = dataFormula["type"]
            coefs = np.array(dataFormula["coefficients"].split()).astype(float)            
            if dispType == 4:
                self.nFunc = lambda wl: _DispersionFunc4(wl, coefs)
            elif dispType == 2:
                self.nFunc = lambda wl: _DispersionFunc2(wl, coefs)
            elif dispType == 1:
                self.nFunc = lambda wl: _DispersionFunc1(wl, coefs)
            else:
                raise NotImplemented()
            self.isFormula = True
            self.wlRange = wlRange
                
        if "DATA" in self.rawData and type(self.rawData["DATA"]) == list:
            coefs = np.array(self.rawData["DATA"][0]["coefficients"].split()).astype(float)
            if self.rawData["DATA"][0]["type"] == "formula 1": 
                self.nFunc = lambda wl: _DisperisonFuncFormula1(wl, coefs)
            elif self.rawData["DATA"][0]["type"] == "formula 2": 
                self.nFunc = lambda wl: _DisperisonFuncFormula2(wl, coefs)
            elif self.rawData["DATA"][0]["type"] == "formula 5":
                self.nFunc = lambda wl: _DisperisonFuncFormula5(wl, coefs)
            else:
                raise NotImplementedError()




#===============================================================================
# Dispersion functions
#===============================================================================


def _DispersionFunc4(wl, coefs):
    wl = wl * 1e6
    n2 = coefs[0]

    for i in range(1, len(coefs), 4):
        if i + 2 < len(coefs):
            n2 += coefs[i] * wl ** coefs[i + 1] / \
                (wl ** 2.0 - coefs[i + 2] ** coefs[i + 3])
        else:
            n2 += coefs[i] * wl ** coefs[i + 1]
    return np.sqrt(n2)

def _DispersionFunc1(wl, coefs):
    wl = wl * 1e6

    n2MinOne = coefs[0]
    for i in range(1, len(coefs), 2):
        n2MinOne += coefs[i] * (wl ** 2.0) / (wl ** 2.0 - coefs[i + 1] ** 2.0)
    return np.sqrt(n2MinOne + 1.0)


def _DispersionFunc2(wl, coefs):
    wl = wl * 1e6

    n2MinOne = coefs[0]
    for i in range(1, len(coefs), 2):
        n2MinOne += coefs[i] * (wl ** 2.0) / (wl ** 2.0 - coefs[i + 1])
    return np.sqrt(n2MinOne + 1.0)

def _DisperisonFuncFormula1(wl, coefs):
    wl = wl * 1e6
    n2 = coefs[0] + 1.0
    for i in range(1, len(coefs), 2):
        n2 += coefs[i] * wl ** 2.0 / (wl ** 2.0 - coefs[i + 1] ** 2.0)
    return np.sqrt(n2)

def _DisperisonFuncFormula2(wl, coefs):
    wl = wl * 1e6
    n2 = coefs[0] + 1.0
    for i in range(1, len(coefs), 2):
        n2 += coefs[i] * wl ** 2.0 / (wl ** 2.0 - coefs[i + 1])
    return np.sqrt(n2)

def _DisperisonFuncFormula5(wl, coefs):
    wl = wl * 1e6
    n = coefs[0]
    for i in range(1, len(coefs), 2):
        n += coefs[i] * (wl ** coefs[i + 1])
    return n

if __name__ == '__main__':
    test = Material(r"special/TiO2 S-1-13")
    print(test(np.linspace(400e-9, 600e-9, 5)))
    
    


from  NonlinearTMM import _SecondOrderNLTMMCython  # @UnresolvedImport

__all__ = ["SecondOrderNLTMM"]
    
SecondOrderNLTMM = _SecondOrderNLTMMCython.SecondOrderNLTMM
    
if __name__ == "__main__":
    pass


import pytest
import numpy as np
from LabPy import Constants
from NonlinearTMM import SecondOrderNLTMM, Material

def SpdcPowerQuantum(wlP1, wlP2, betaP1, betaP2, nF, chi2, crystalL, pwrP1, dwl, solidAngleSpdc, deltaThetaSpdc):
    wlGen = Constants.OmegaToWl(Constants.WlToOmega(wlP1) - Constants.WlToOmega(wlP2))
    betaGen = wlGen * (betaP1 / wlP1 - betaP2 / wlP2);
    
    # Omegas
    omegaP1 = Constants.WlToOmega(wlP1)
    omegaP2 = Constants.WlToOmega(wlP2)
    omegaGen = Constants.WlToOmega(wlGen)
    
    # Refractive indices
    nP1 = nF(wlP1)
    nP2 = nF(wlP2)
    nGen = nF(wlGen)
    
    # Wave vectors
    k0P1 = 2.0 * np.pi / wlP1
    k0P2 = 2.0 * np.pi / wlP2
    k0Gen = 2.0 * np.pi / wlGen
    kP1 = k0P1 * nP1
    kP2 = k0P2 * nP2
    kGen = k0Gen * nGen
    
    # Pump 1
    kxP1 = betaP1 * k0P1
    kzP1 = np.sqrt(kP1 ** 2.0 - kxP1 ** 2.0)
    
    # Pump 2 / vacuum fluctuations
    kxP2 = k0P2 * betaP2
    kzP2 = np.sqrt(kP2 ** 2.0 - kxP2 ** 2.0)
    
    # Gen
    kxGen = betaGen * k0Gen
    kzGenSqr = kGen ** 2.0 - kxGen ** 2.0
    kzGen = np.sqrt(kzGenSqr)
    
    # dk
    dkz = kzP1 - kzP2 - kzGen
    #print("wlGen, betaGen, dkz", 1e9 * wlGen, betaGen, dkz)
        
    res = dwl * solidAngleSpdc / deltaThetaSpdc * \
        Constants.hp * chi2 ** 2.0 / (8.0 * np.pi ** 4 * Constants.eps0) * \
        omegaGen ** 6 * omegaP1 * omegaP2 ** 2 / (Constants.c ** 8) * pwrP1 * \
        1.0 / (kzGen * kzP1 * kP2) * \
        np.sin(0.5 * dkz * crystalL) ** 2 / (0.5 * dkz) ** 2
        
    return res

@pytest.fixture(params = [0.0, 0.3])
def betaP1(request):
    return request.param

@pytest.fixture(params = [0.1e-3, 1e-3])
def crystalD(request):
    return request.param

@pytest.fixture(params = [0.1, 1.0])
def pwrP1(request):
    return request.param

@pytest.fixture(params = [1.61681049944, 1.5])
def n1(request):
    return request.param

@pytest.fixture(params = [1.6913720951, 1.52])
def n2(request):
    return request.param

@pytest.fixture(params = ["planewave", "gaussian"])
def waveP1Type(request):
    return request.param

@pytest.fixture()
def tmmParams(crystalD, pwrP1, n1, n2):
    # Define params
    wlP1 = 450e-9
    wlP2 = 800e-9
    polP1 = "s"
    polP2 = "s"
    polGen = "s"
    w0 = 1e-3
    Ly = 1e-3
    I0 = pwrP1  / (w0 * Ly)
    
    deltaThetaSpdc = np.radians(0.5)
    solidAngleSpdc = 7.61543549467e-05
    deltaWlSpdc = 2.5e-9
    chi2 = 1.4761823412e-12
    
    # Define materials
    wlsCrystal = np.array([100e-9, 1200.0001e-9])
    nsCrystal = np.array([n1, n2], dtype = complex)
    prism = Material(wlsCrystal, nsCrystal)
    crystal = Material(wlsCrystal, nsCrystal)
    dielectric = Material(wlsCrystal, nsCrystal)
    crystal.chi2.Update(chi222 = chi2, chi111 = chi2, chi333 = chi2)
    
    # Init SecondOrderNLTMM
    tmm = SecondOrderNLTMM(mode = "spdc", deltaWlSpdc = deltaWlSpdc, \
                           solidAngleSpdc = solidAngleSpdc, deltaThetaSpdc = deltaThetaSpdc)
    tmm.P1.SetParams(wl = wlP1, pol = polP1, I0 = I0)
    tmm.P2.SetParams(wl = wlP2, pol = polP2, overrideE0 = True, E0 = 1.0)
    tmm.Gen.SetParams(pol = polGen)
    
    # Add layers
    tmm.AddLayer(float("inf"), prism)
    tmm.AddLayer(crystalD, crystal)
    tmm.AddLayer(float("inf"), dielectric)
    
    return tmm, prism, chi2, w0, Ly
    
    
def test_SPDCIntensity(tmmParams, betaP1):
    tmm, nF, chi2, w0, Ly = tmmParams
    betasP2 = np.linspace(0.0, 0.9, 10000)
    betasP1 = np.ones_like(betasP2) * betaP1
    pwrP1 = tmm.P1.I0
    crystalD = tmm.P1.layers[1].d
    
    # Sweep over beta = sin(th) * n_prism
    sr = tmm.Sweep("beta", betasP1, betasP2)
    quantumSpdc = SpdcPowerQuantum(tmm.P1.wl, tmm.P2.wl, betaP1, betasP2, \
        lambda wl: nF.GetN(wl).real, chi2, crystalD, pwrP1, tmm.deltaWlSpdc, tmm.solidAngleSpdc, tmm.deltaThetaSpdc)
    coef = 1.0 / np.cos(np.arcsin(betasP1 / nF.GetN(tmm.P1.wl).real)) 
    
    np.testing.assert_allclose(quantumSpdc, sr.Gen.It * coef, rtol = 1e-6)

def test_SPDCPwrs(tmmParams, waveP1Type, betaP1):
    tmm, nF, chi2, w0, Ly = tmmParams 
    betasP2Test = np.linspace(0.0, 0.9, 10000)
    betasP1Test = np.ones_like(betasP2Test) * betaP1
    
    # Waves
    pwrP1 = tmm.P1.I0 * w0 * Ly
    waveP1Params = {"waveType": waveP1Type, "pwr": pwrP1, "w0": w0, "Ly": Ly, "dynamicMaxXCoef": 10, "nPointsInteg": 100, "maxPhi": np.radians(10.0)}
    waveP2Params = {"waveType": "spdc", "w0": w0, "Ly": Ly, "overrideE0": True, "nPointsInteg": 100}
    tmm.P1.wave.SetParams(**waveP1Params)
    tmm.P2.wave.SetParams(**waveP2Params)

    # Find maximum signal
    srTest = tmm.Sweep("beta", betasP1Test, betasP2Test)
    index = np.argmax(srTest.Gen.It)
    betaP2 = betasP2Test[index]
    
    # Manual integration
    # --------------------------------------------------------------------------
    
    # Find betaP2 integration range
    wlGen = srTest.wlsGen[0]
    n0Gen = nF.GetN(wlGen).real
    maxBetaGen = srTest.betasGen[index]
    kz0Gen = (2.0 * np.pi / wlGen) * np.sqrt(n0Gen ** 2 - maxBetaGen ** 2)
    deltaKxP2 = kz0Gen / (2.0 * n0Gen) * tmm.deltaThetaSpdc
    deltaBetaP2 = deltaKxP2 / (2.0 * np.pi / tmm.P2.wl)
    
    # Integrate
    betasP2Int = np.linspace(betaP2 - deltaBetaP2, betaP2 + deltaBetaP2, 100)
    betasP1Int = np.ones_like(betasP2Int) * betaP1
    kxsP2Int = betasP2Int * (2.0 * np.pi / tmm.P2.wl)
    srInt = tmm.Sweep("beta", betasP1Int, betasP2Int)
    
    # Int power
    pwrSPDC = np.trapz(srInt.Gen.It, kxsP2Int)
    
    # Wave integration
    # --------------------------------------------------------------------------
    
    tmm.P1.beta = betaP1
    tmm.P2.beta = betaP2
    pwrSpdcW = tmm.WaveGetPowerFlows(2)[0]
    beamArea = tmm.P1.wave.beamArea
    
    # Testing
    # --------------------------------------------------------------------------
    np.testing.assert_allclose(beamArea * pwrSPDC, pwrSpdcW, rtol = 1e-2)
    
    
    




import numpy as np
import pylab as plt
from NonlinearTMM import SecondOrderNLTMM, Material

if __name__ == "__main__":
    # Define params
    wlP1 = 1000e-9
    wlP2 = 1000e-9
    polP1 = "s"
    polP2 = "s"
    polGen = "s"
    I0P1 = 1.0
    I0P2 = I0P1
    betas = np.linspace(0.0, 0.99, 10000)
    crystalD = 1000e-6
    
    # Define materials
    wlsCrystal = np.array([400e-9, 1100e-9])
    nsCrystal = np.array([1.54, 1.53], dtype = complex)
    prism = Material.Static(1.0)
    crystal = Material(wlsCrystal, nsCrystal)
    dielectric = Material.Static(1.0)
    crystal.chi2.Update(d22 = 1e-12)
    
    # Init SecondOrderNLTMM
    tmm = SecondOrderNLTMM()
    tmm.P1.SetParams(wl = 1000e-9, pol = "s", beta = 0.2, I0 = 1.0)
    tmm.P2.SetParams(wl = 1000e-9, pol = "s", beta = 0.2, I0 = 1.0)
    tmm.Gen.SetParams(pol = "s")
    
    # Add layers
    tmm.AddLayer(float("inf"), prism)
    tmm.AddLayer(crystalD, crystal)
    tmm.AddLayer(float("inf"), dielectric)
    
    # Sweep over beta = sin(th) * n_prism
    sr = tmm.Sweep("beta", betas, betas)
    
    # Plot generated reflection and transmission
    plt.title("SHG generation from crystal (d = %.0f $\mu m$)" % (1e6 * crystalD))
    plt.plot(betas, sr.Gen.Ir, label = "R")
    plt.plot(betas, sr.Gen.It, label = "T")
    plt.legend()
    plt.xlabel(r"$\beta$")
    plt.ylabel(r"($W / m^{2}$)")
        
    plt.show()
    
    


import pytest
import numpy as np
from NonlinearTMM import TMM, Material

@pytest.fixture()
def tmm():
    wl = 532e-9
    metalD = 60e-9

    prism = Material.Static(1.5)
    metal = Material.Static(0.054007 + 3.4290j)
    dielectric = Material.Static(1.0)
    
    res = TMM(wl = wl)
    res.AddLayer(float("inf"), prism)
    res.AddLayer(metalD, metal)
    res.AddLayer(float("inf"), dielectric)
    
    return res

@pytest.fixture(params = ["p", "s"])
def pol(request):
    return request.param

@pytest.fixture(params = ["tukey", "gaussian", "planewave"])
def waveType(request):
    return request.param


def test_WaveSweep(tmm, pol, waveType):
    betas = np.linspace(0.0, 0.5, 15)
    enhLayer = 2
    enhZ = 0.0
    Ly = 1e-3
    pwr = 10e-3
    w0 = 500e-6
    waveParams = {"waveType": waveType, "pwr": pwr, "w0": w0, "Ly": Ly, "nPointsInteg": 300, "dynamicMaxXCoef": 10.0}
    
    tmm.SetParams(pol = pol)
    tmm.wave.SetParams(**waveParams)

    # Ordinary sweep
    tmm.I0 = pwr / (w0 * Ly)
    sr = tmm.Sweep("beta", betas, layerNr = enhLayer, layerZ = enhZ, outEnh = True)
    
    # Waves sweep
    srW = tmm.WaveSweep("beta", betas, layerNr = enhLayer, layerZ = enhZ, outEnh = True)    
    
    
    # Test incident power
    # Convert intensity to power
    pwrIncident = sr.Ii * srW.beamArea
    np.testing.assert_allclose(srW.Pi, pwrIncident, atol = 1e-4)
        
    # Test refl, transmission, enhancment
    np.testing.assert_allclose(srW.Pr / srW.Pi, sr.Ir / sr.Ii, atol = 1e-4)
    np.testing.assert_allclose(srW.Pt / srW.Pi, sr.It / sr.Ii, atol = 1e-4)
    np.testing.assert_allclose(srW.enh, sr.enh, atol = 5e-4)
    


import numpy as np
from LabPy import Core, Constants
from time import time

__all__ = ["PlaneWave", \
           "GaussianWave",
           "GaussianWaveFFT",
           "TukeyWaveFFT",
           "TmmForWaves",
           "SecondOrderNLTmmForWaves"]

#===============================================================================
# PlaneWave
#===============================================================================

class PlaneWave(Core.ParamsBaseClass):
    
    def __init__(self, params = [], **kwargs):
        paramsThis = ["pwr", "overrideE0", "w0", "n", "Ly"]
        self.overrideE0 = None #E0 specified in vacuum, not in first layer
        self.Ly = None
        super(PlaneWave, self).__init__(params + paramsThis, **kwargs)

    def Solve(self, wl, th0, **kwargs):
        self.wl = wl
        self.th0 = th0
        self.SetParams(**kwargs)
        
        self._Precalc()
        self._Solve()

    def _Precalc(self): 
        if self.overrideE0 is None:
            self.I0 = self.pwr / (self.Ly * self.w0) 
            self.E0 = np.sqrt(2.0 * Constants.mu0 * Constants.c * self.I0) 
            # E0 in vacuum, not in 1st layer
            print("E0 vacuum", self.E0)
        else:
            self.E0 = self.overrideE0
            
        self.k0 = 2.0 * np.pi / self.wl
        self.k = self.k0 * self.n(self.wl).real
        print("_Precalc", self.E0)
        
    def _Solve(self):
        self.phis = np.array([0.0])
        self.kxs = np.array([self.k * np.sin(self.th0)])
        self.kzs = np.array([self.k * np.cos(self.th0)])
        self.expansionCoefsPhi = np.array([self.E0])
        self.expansionCoefsKx = np.array([self.E0])
        self.betas = (self.kxs / self.k0).real
   
       
#===============================================================================
# GaussianWave
#===============================================================================

class GaussianWave(PlaneWave):
    def __init__(self, params = [], **kwargs):
        paramsThis = ["integCriteria", "nPointsInteg"]
        self.nPointsInteg = 30
        self.integCriteria = 1e-3
        super(GaussianWave, self).__init__(params + paramsThis, **kwargs)
    
    def _Solve(self):
        self._phiLim  = np.arcsin(2.0 / self.w0 * np.sqrt(-np.log(self.integCriteria)) / self.k)
        self.phis = np.linspace(-self._phiLim, self._phiLim, self.nPointsInteg)

        if np.max(abs(self.phis + self.th0)) > np.pi / 2.0:
            raise ValueError("Gaussian wave requires backward propagating waves!")
        
        kzPs, kxPs = np.cos(self.phis) * self.k, np.sin(self.phis) * self.k
        self.kxs = kxPs * np.cos(self.th0) + kzPs * np.sin(self.th0)
        self.kzs = -kxPs * np.sin(self.th0) + kzPs * np.cos(self.th0)
        profileSpectrum = self.E0 * 1.0 / 2.0 / np.sqrt(np.pi) * self.w0 * \
            np.exp(-(kxPs) ** 2.0 * self.w0 ** 2.0 / 4.0)
        self.expansionCoefsPhi = profileSpectrum * np.cos(self.phis) * self.k
        self.betas = (self.kxs / self.k0).real

#===============================================================================
# WaveFFT
#===============================================================================

class WaveFFT(PlaneWave):
    def __init__(self, params = [], **kwargs):
        paramsThis = ["nPointsInteg", "maxPhi", "integCriteria"]
        self.nPointsInteg = 30
        self.maxPhi = np.radians(0.5)
        self.integCriteria = 1e-3
        PlaneWave.__init__(self, params + paramsThis, **kwargs)
    
    def _FieldProfile(self, xs):
        raise NotImplementedError()
        
    def _Solve(self, maxPhiForce = None):
        maxKxp = np.sin(self.maxPhi if maxPhiForce is None else maxPhiForce) * self.k
        dx = np.pi / maxKxp 
        xs = np.arange(-0.5 * dx * self.nPointsInteg, 0.5 * dx * self.nPointsInteg, dx)
        
        fieldProfile = self._FieldProfile(xs)
        fieldProfileSpectrum = (xs[1] - xs[0]) / (2.0 * np.pi) * np.fft.fftshift(np.fft.fft(np.fft.ifftshift(fieldProfile)))
        kxPs = 2.0 * np.pi * np.fft.fftshift(np.fft.fftfreq(len(fieldProfile), xs[1] - xs[0]))
        self.phis = np.arcsin(kxPs / self.k)

        if maxPhiForce is None and self.integCriteria is not None:
            # It is first iteration, need to check if it is possible to reduce the range
            # by using the integCriteria
            cumSpectra = np.cumsum(abs(fieldProfileSpectrum))
            cumSpectra /= np.max(cumSpectra)
            index = np.argmax(cumSpectra > self.integCriteria)
            
            if index > 0:
                newMaxPhi = abs(self.phis[max(index - 5, 0)])
                print("newMaxPhi", np.degrees(newMaxPhi), cumSpectra[index])
                self._Solve(newMaxPhi)
                return

        if np.max(abs(self.phis + self.th0)) > np.pi / 2.0:
            raise ValueError("Requires backward propagating waves!")
        
        kzPs = np.cos(self.phis) * self.k
        self.kxs = kxPs * np.cos(self.th0) + kzPs * np.sin(self.th0)
        self.kzs = -kxPs * np.sin(self.th0) + kzPs * np.cos(self.th0)
        
        self.expansionCoefsPhi = fieldProfileSpectrum * np.cos(self.phis) * self.k
        self.expansionCoefsKx = fieldProfileSpectrum / np.cos(self.th0)
        self.betas = (self.kxs / self.k0).real
    
#===============================================================================
# GaussianWaveFFT
#===============================================================================

class GaussianWaveFFT(WaveFFT):
    def __init__(self, params = [], **kwargs):
        paramsThis = []
        WaveFFT.__init__(self, params + paramsThis, **kwargs)
    
    def _FieldProfile(self, xs):
        res = self.E0 * np.exp(- xs ** 2.0 / self.w0 ** 2.0)
        return res

#===============================================================================
# TukeyWaveFFT
#===============================================================================

class TukeyWaveFFT(WaveFFT):
    def __init__(self, params = [], **kwargs):
        paramsThis = ["a"]
        self.a = 0.5
        WaveFFT.__init__(self, params + paramsThis, **kwargs)
    
    @staticmethod
    def TukeyFunc(xs, w0, a):
        lowering = 0.5 * np.cos(np.pi / (w0 * (1.0 - a)) * (xs - 0.5 * a * w0)) + 0.5
        raising = 0.5 * np.cos(np.pi / (w0 * (1.0 - a)) * (xs + 0.5 * (2.0 - a) * w0) + np.pi) + 0.5
        
        res = np.ones_like(xs, dtype = float)
        res = np.where(xs > 0.5 * a * w0, lowering, res)
        res = np.where(xs < -0.5 * a * w0, raising, res)
        res[xs > 0.5 * w0 * (2.0 - a)] = 0.0
        res[xs < -0.5 * w0 * (2.0 - a)] = 0.0
        return res
    
    def _FieldProfile(self, xs):
        res = self.E0 * self.TukeyFunc(xs, self.w0, self.a)
        return res
        
#===============================================================================
# WavesIntegrator1D
#===============================================================================

def WavesIntegrator1D(wave, fieldsFunc):
    resE, resH = None, None
    lastE, lastH = None, None
    dPhis = wave.phis[1:] - wave.phis[:-1] 

    for i in range(len(wave.betas)):
        fieldsE, fieldsH = fieldsFunc(wave, i)
        
        if len(wave.betas) == 1:
            return fieldsE, fieldsH
        
        if i == 0:
            resE = np.zeros_like(fieldsE)
            resH = np.zeros_like(fieldsH)    
        else:
            resE += dPhis[i - 1] * (fieldsE + lastE) / 2.0
            resH += dPhis[i - 1] * (fieldsH + lastH) / 2.0
        
        lastE = fieldsE
        lastH = fieldsH
    return resE, resH

def WavesIntegrator2D(wave1, wave2, fieldsFunc):
    resE, resH = None, None
    lastE, lastH = None, None
    dPhis = wave1.phis[1:] - wave1.phis[:-1] 

    for i in range(len(wave1.betas)):
        fieldsFunc2 = lambda w, nr: fieldsFunc(wave1, i, w, nr)
        fieldsE, fieldsH = WavesIntegrator1D(wave2, fieldsFunc2)
        
        if len(wave1.betas) == 1:
            return fieldsE, fieldsH
        
        if i == 0:
            resE = np.zeros_like(fieldsE)
            resH = np.zeros_like(fieldsH)    
        else:
            resE += dPhis[i - 1] * (fieldsE + lastE) / 2.0
            resH += dPhis[i - 1] * (fieldsH + lastH) / 2.0
        
        lastE = fieldsE
        lastH = fieldsH
        
    return resE, resH

#===============================================================================
# TmmForWavePackets
#===============================================================================

class TmmForWaves(Core.ParamsBaseClass):

    def __init__(self, tmm, wave, **kwargs):
        self._params = ["th0", "pol"]
        self.wave = wave 
        self.tmm = tmm
        self.th0 = None
        self.pol = None
        super(TmmForWaves, self).__init__(**kwargs)
    
    def CalcFields2d(self, xs, ys, **kwargs):
        
        def IntegFunc(wave, nr):
            self.tmm.SetParam(beta = wave.betas[nr])
            E, H = self.tmm.CalcFields2D(xs, ys, self.pol)
            E *= wave.expansionCoefsPhi[nr]
            H *= wave.expansionCoefsPhi[nr]
            return E, H
        
        self.SetParams(**kwargs)
        wl = self.tmm.GetParam("wl").real
        nPrism = self.tmm._materialsCache[0][1]
        self.wave.Solve(wl, self.th0, n = nPrism)

        ESum, HSum = WavesIntegrator1D(self.wave, IntegFunc)
        return ESum, HSum

#===============================================================================
# SecondOrderNLTmmForWaves
#===============================================================================

class SecondOrderNLTmmForWaves(Core.ParamsBaseClass):
    
    def __init__(self, tmmNL, wave1, wave2, **kwargs):
        self._params = ["thP1", "thP2"]
        self.tmmNL = tmmNL
        self.wave1 = wave1
        self.wave2 = wave2
        self.thP1 = None
        self.thP2 = None
        
        super(SecondOrderNLTmmForWaves, self).__init__(**kwargs)
    
    # Powers (porks only in linear media!)

    def CalcPowersP1(self, layerNr, x0, x1, z, **kwargs):
        self.SetParams(**kwargs)
        res = self._CalcPowersPump(self.tmmNL.tmmP1, self.thP1, self.wave1, layerNr, x0, x1, z)
        return res

    def CalcPowersP2(self, x0, x1, z, **kwargs):
        self.SetParams(**kwargs)
        res = self._CalcPowersPump(self.tmmNL.tmmP2, self.thP2, self.wave2, x0, x1, z)
        return res
    
    def CalcPowersGen(self, layerNr, x0, x1, z, **kwargs):
        self.SetParams(**kwargs)
        
        # Solve wave
        starttime = time()
        self.wave1.Solve(self.tmmNL.wlP1, self.thP1, n = self.tmmNL.layers[0][1])
        self.wave2.Solve(self.tmmNL.wlP2, self.thP2, n = self.tmmNL.layers[0][1])
        print("wave.Solve", time() - starttime)
        
        # Init variables
        starttime = time()
        kxs = np.zeros((len(self.wave1.kxs) * len(self.wave2.kxs)))
        UsF = np.zeros_like(kxs, dtype = complex)
        UsB = np.zeros_like(kxs, dtype = complex)
        kzsF = np.zeros_like(kxs, dtype = complex)
        kzsB = np.zeros_like(kxs, dtype = complex)
        print("init variables", time() - starttime)
        
        # Solve system for every beta
        starttime = time()
        for nr1 in range(len(self.wave1.kxs)):
            for nr2 in range(len(self.wave2.kxs)): 
                self.tmmNL.SetParams(betaP1 = self.wave1.betas[nr1],
                                     betaP2 = self.wave2.betas[nr2],
                                     overrideE0P1 = self.wave1.expansionCoefsKx[nr1],
                                     overrideE0P2 = self.wave2.expansionCoefsKx[nr2])
                self.tmmNL.Solve()
                
                nr = nr1 * len(self.wave2.kxs) + nr2
                kxs[nr] = self.tmmNL.betaGen * 2.0 * np.pi / self.tmmNL.wlGen
                U = self.tmmNL.tmmGen.layers[layerNr].hw.GetMainFields(0.0)
                UsF[nr], UsB[nr] = U[0, 0], U[1, 0]
                kzsF[nr] = self.tmmNL.tmmGen.layers[layerNr].hw.kzF
                kzsB[nr] = self.tmmNL.tmmGen.layers[layerNr].hw.kzB
        print("solve system", time() - starttime)
                
        # Sort by kx
        starttime = time()
        sortP = kxs.argsort()
        kxs = kxs[sortP]
        UsF = UsF[sortP]
        UsB = UsB[sortP]
        kzsF = kzsF[sortP]
        kzsB = kzsB[sortP]
        print("sorting", time() - starttime)
        
        Ly = self.wave1.Ly
        if self.wave2.Ly != Ly:
            raise ValueError("All input waves must have same Ly")
        
        PF = self._IntegrateWavePower(self.tmmNL.tmmGen, UsF, kxs, kzsF, layerNr, x0, x1, z, Ly)
        PB = -self._IntegrateWavePower(self.tmmNL.tmmGen, UsB, kxs, kzsB, layerNr, x0, x1, z, Ly)
        return PF, PB
        
        
    # Fields
    
    def CalcFields2dP1(self, zs, xs, **kwargs):
        self.SetParams(**kwargs)
        res = self._CalcFields2dPump(self.tmmNL.tmmP1, self.thP1, self.wave1, zs, xs)
        return res

    def CalcFields2dP2(self, zs, xs, **kwargs):
        self.SetParams(**kwargs)
        res = self._CalcFields2dPump(self.tmmNL.tmmP2, self.thP2, self.wave2, zs, xs)
        return res        
    
    def CalcFields2dGen(self, zs, xs, **kwargs):
        def IntegFunc(wave1, nr1, wave2, nr2):
            self.tmmNL.SetParams(betaP1 = wave1.betas[nr1],
                                 betaP2 = wave2.betas[nr2],
                                 overrideE0P1 = wave1.expansionCoefsPhi[nr1],
                                 overrideE0P2 = wave2.expansionCoefsPhi[nr2])
            self.tmmNL.Solve()
            E, H = self.tmmNL.tmmGen.GetFields2D(zs, xs)
            return E, H

        self.SetParams(**kwargs)
        self.wave1.Solve(self.tmmNL.wlP1, self.thP1, n = self.tmmNL.layers[0][1])
        self.wave2.Solve(self.tmmNL.wlP2, self.thP2, n = self.tmmNL.layers[0][1])
        ESum, HSum = WavesIntegrator2D(self.wave1, self.wave2, IntegFunc)
        return ESum, HSum

    # Private methods

    def _CalcFields2dPump(self, tmmP, thP, wave, zs, xs):
        def IntegFunc(wave, nr):
            tmmP.SetParams(beta = wave.betas[nr], overrideE0 = wave.expansionCoefsPhi[nr])
            tmmP.Solve()
            E, H = tmmP.GetFields2D(zs, xs)
            return E, H
        
        wave.Solve(tmmP.wl, thP, n = tmmP.layers[0].n)
        ESum, HSum = WavesIntegrator1D(wave, IntegFunc)
        return ESum, HSum
    
    def _CalcPowersPump(self, tmmP, thP, wave, layerNr, x0, x1, z):
        # Solve wave
        starttime = time()
        wave.Solve(tmmP.wl, thP, n = tmmP.layers[0].n)
        print("wave.Solve", time() - starttime)
        
        # Init variables
        starttime = time()
        kxs = wave.kxs
        UsF = np.zeros_like(kxs, dtype = complex)
        UsB = np.zeros_like(kxs, dtype = complex)
        kzsF = np.zeros_like(kxs, dtype = complex)
        kzsB = np.zeros_like(kxs, dtype = complex)
        print("init variables", time() - starttime)
        
        # Solve system for every beta
        starttime = time()
        for nr in range(len(kxs)): 
            tmmP.SetParams(beta = wave.betas[nr], overrideE0 = wave.expansionCoefsKx[nr])
            tmmP.Solve()
            U = tmmP.layers[layerNr].hw.GetMainFields(0.0)
            UsF[nr], UsB[nr] = U[0, 0], U[1, 0]
            kzsF[nr] = tmmP.layers[layerNr].hw.kzF
            kzsB[nr] = tmmP.layers[layerNr].hw.kzB
        print("solve system", time() - starttime)
        
        PF = self._IntegrateWavePower(tmmP, UsF, kxs, kzsF, layerNr, x0, x1, z, wave.Ly)
        PB = -self._IntegrateWavePower(tmmP, UsB, kxs, kzsB, layerNr, x0, x1, z, wave.Ly)
        return PF, PB
            
    def _Integrate(self, xs, values):
        dxs = xs[1:] - xs[:-1]
        res = np.sum(dxs * 0.5 * (values[1:] + values[:-1]))
        return res
    
    def _IntegrateWavePower(self, tmm, Us, kxs, kzs, layerNr, x0, x1, z, Ly):
        # Integrate wave
        starttime = time()    
        
        if len(kxs) == 1:
            # Plane wave
            integ2d = Us[0] * np.conj(Us[0]) * (x1 - x0) * kzs[0]
        else:
            kxP = kxs
            kzP = kzs
            
            # Internal integral
            integ1ds = np.zeros_like(kxs, dtype = complex)
            for i, (kx, kz) in enumerate(zip(kxs, kzs)):
                # Fx
                dk = kx - kxP
                dk[i] = 1.0
                Fx = -1.0j / dk * (np.exp(1.0j * dk * x1) - np.exp(1.0j * dk * x0))
                Fx[i] = x1 - x0
                dk[i] = 0.0
                
                kzPart = kz if tmm.pol == "p" else kzP
                integValues = np.conj(Us) * Fx * kzPart * np.exp(1.0j * (kz - kzP) * z)     
                integ1ds[i] = self._Integrate(kxs, integValues)
 
            # Outer integral
            integ2d = self._Integrate(kxs,  Us * integ1ds)
            
        omega = Constants.WlToOmega(tmm.wl)
        if tmm.pol == "p":
            P = Ly / (2.0 * omega * Constants.eps0) * (1.0 / tmm.layers[layerNr].eps * integ2d).real
        elif tmm.pol == "s":
            P = Ly / (2.0 * omega * Constants.mu0) * (integ2d).real
        else:
            raise NotImplementedError()
        print("integration", time() - starttime)
        
        return P
    
if __name__ == "__main__":
    pass


from NonlinearTMM import _SecondOrderNLTMMCython  # @UnresolvedImport

__all__ = ["TMM",
           "NonlinearTMM"]
    
TMM = _SecondOrderNLTMMCython.NonlinearTMM
NonlinearTMM = _SecondOrderNLTMMCython.NonlinearTMM

if __name__ == "__main__":
    pass


__version__ = "1.3.4"

from ._Material import *
from ._NonlinearTMM import *
from ._SecondOrderNLTMM import *

# Helper classes
from ._SecondOrderNLTMMCython import _Chi2Tensor  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _Wave  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _Intensities  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _SweepResultNonlinearTMM  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _WaveSweepResultNonlinearTMM  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _FieldsZ  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _FieldsZX  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _HomogeneousWave  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _NonlinearLayer  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _SecondOrderNLIntensities  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _SweepResultSecondOrderNLTMM  # @UnresolvedImport
from ._SecondOrderNLTMMCython import _WaveSweepResultSecondOrderNLTMM  # @UnresolvedImport

if __name__ == "__main__":
    pass


import numpy as np
from  NonlinearTMM import _SecondOrderNLTMMCython  # @UnresolvedImport

__all__ = ["Material"]

class Material(_SecondOrderNLTMMCython.Material):# @UndefinedVariable
    __doc__ = _SecondOrderNLTMMCython.Material.__doc__
    
    @staticmethod
    def Static(n):
        """Helper method to make material with constant refractive index.
        
        Parameters
        ----------
        n : float or complex
            Constant value for refractive index
            
        Returns
        -------
        None
        
        Examples
        --------
        >>> mat = Material.Static(1.5)
        >>> mat.GetN(532e-9)
        1.5
        
        """
        wls = np.array([-1.0, 1.0])
        ns = np.array([n, n], dtype = complex)
        res = Material(wls, ns)
        return res
    
    @staticmethod
    def FromLabPy(materialLabPy):
        if materialLabPy.materialFile == "Static":
            wls = np.array([-1.0, 1.0])
            n = materialLabPy.n + 1.0j * (materialLabPy.k + materialLabPy.kAdditional)
            ns = np.array([n, n], dtype = complex)
        elif materialLabPy.isFormula:
            wls = np.ascontiguousarray(np.linspace(materialLabPy.wlRange[0], materialLabPy.wlRange[1], 500))
            ns = np.ascontiguousarray(materialLabPy(wls))
        else:
            wls = np.ascontiguousarray(materialLabPy.wlExp)
            if materialLabPy.kExp is None:
                ns = np.ascontiguousarray(materialLabPy.nExp, dtype = complex)
            else:
                ns = np.ascontiguousarray(materialLabPy.nExp + 1.0j * materialLabPy.kExp)
            ns += 1.0j * materialLabPy.kAdditional
        res = Material(wls, ns)
        res._materialLabPy = materialLabPy
        return res
   
if __name__ == "__main__":
    pass


"""This module contains physical constants and functions to convert between 
different units.

Attributes:
    c (float): speed of light (m/s)
    h (float): Planck's constant (J*s)
    hp (float): Reduced Planck's constant (J*s)
    qe (float): Elementary charge (C)
    eps0 (float): Vacuum permittivity (F/m)
    mu0 (float): Vacuum permeability (A/m)

"""
import math
import numpy as np
from LabPy import Core  # @UnresolvedImport

#===============================================================================
# Definition of constants
#===============================================================================


c = 299792458.0
h = 6.62606957e-34
hp = h / (2.0 * math.pi)
qe = 1.602176565e-19 
eps0 = 8.854187817e-12
mu0 = 1.2566370614e-6
pi = math.pi
kB = 1.38064852e-23


#===============================================================================
# Wavelength to ... conversion functions
#===============================================================================


def WlToJoule(wl):
    """Converts wavelength (m) to photon energy (J).
    
    Args:
        wl (float): wavelength (m)
        
    Returns: 
        float: energy (J)
    
    """
    E = h * c / wl
    return E

def WlToEv(wl):
    """Converts wavelength (m) to energy (eV).
    
    Args:
        wl (float): wavelength
    
    Returns: 
        float: energy (eV)
    
    """
    ev = h * c / (wl * qe)
    return ev

def WlToFreq(wl):
    """Converts wavelength to frequency (Hz).
    
    Args:
    wl (float): wavelength (m)
    
    Returns: 
        float: frequency (Hz)
    
    """
    freq = c / wl
    return freq 


def WlToOmega(wl):
    """Converts wavelength to angular frequency (rad/s).
    
    Args:
        wl (float): wavelength (m)
    
    Returns:
        float: angular frequency (rad/s)
    
    """
    omega = 2.0 * math.pi * c / wl
    return omega

#===============================================================================
# ... to wavelength conversion functions
#===============================================================================

def JouleToWl(E):
    """Converts energy of a photon (J) to wavelength (m).
    
    Args:
        E (float): energy (J)
    
    Returns: 
        float: wavelength (nm)
    
    """
    wl = h * c / E
    return wl


def EvToWl(ev):
    """Converts energy (eV) to wavelength (m).
    
    Args:
        ev (float): energy (eV)
    
    Returns: 
        float: wavelength (m)
    
    """
    wl = h * c / (ev * qe)
    return wl


def FreqToWl(freq):
    """Converts frequency (1/s) to wavelength (m).
    
    Args:
        freq (float): frequency (1/s)
        
    Returns: 
        float: wavelength (m)
        
    """
    wl = c / freq
    return wl


def OmegaToWl(omega):
    """Converts angular frequency (rad/s) to wavelength (m).
    
    Args:
        omega (float): angular frequency (rad/s)

    Returns: 
        float: wavelength (m)
    
    """
    wl = 2.0 * math.pi * c / omega
    return wl

#===============================================================================
# Other conversion functions
#===============================================================================
 
def EvToJoule(ev):
    """Converts eV-s to joules.
    
    Args:
        ev (float): Energy in eV
        
    Returns:
        float: Energy in J
        
    """
    
    return ev * qe
 
def JouleToEv(E):
    """Converts joules to eV-s.
    
    Args:
        E (float): Energy in J
        
    Returns: 
        float: Energy in eV
        
    """
    
    return E / qe

def EvToOmega(ev):
    """Converts eV-s to angular frequency.
    
    Args:
        ev (float): Energy in eV
        
    Returns: 
        float: Photon angular frequency (rad/s)
        
    """
    return 2.0 * math.pi * ev * qe / h

def OmegaToEv(omega):
    """Converts angular frequency to eV-s.
    
    Args:
        omega (float): photon angular frequency (rad/s)
        
    Returns: 
        float: Photon energy (eV)
        
    """
    return omega * h / (2.0 * pi * qe)

#===============================================================================
# Distributions
#===============================================================================     
    

def GetDistributionNames():
    res = ["DetlaDistribution", "NormalDistribution", \
                     "LogNormalDistribution", "LogNormalDistributionLocal"]
    return res

    
def Distribution(name, **kwargs):
    if name == "DetlaDistribution":
        return DetlaDistribution(**kwargs)
    elif name == "NormalDistribution":
        return NormalDistribution(**kwargs)
    elif name == "LogNormalDistribution":
        return LogNormalDistribution(**kwargs)
    elif name == "LogNormalDistributionLocal":
        return LogNormalDistributionLocal(**kwargs)
    else:
        raise ValueError("Unknown distribution name")
    
class DetlaDistribution(Core.ParamsBaseClass):
    """This class presents delta function distribution.
    
    Delta function distribution is zero except at x=x0 where it is 1.0.
    
    Kwargs:
        x0=0.0 (float): The position of delta function
    
    """
    
    def __init__(self, **kwargs):
        self._params = ["x0"]
        self.x0 = 0.0
        
        super(DetlaDistribution, self).__init__(**kwargs)
        
    def __call__(self, x):
        """Returns distribution value at position x.
        
        Args:
            x (float): The point were to evaluate the distribution
        
        Returns:
            float: Distribution value at position x
            
        """
        if x == self.x0:
            return 1.0
        else:
            return 0.0
        
    def GetDist(self):
        """Function returns (np.array([x0]), np.array([1.0])).
        
        Returns:
            tuple_of_arrays: Distribution positions and values
            
        """
        return (np.array([self.x0]), np.array([1.0]))
        
class NormalDistribution(Core.ParamsBaseClass):
    """This class presents normal distribution.
    
    Kwargs:
        x0=0.0 (float): mean
        std=1.0 (float): standard deviation
        nPoints=30 (int): number of points in distribution
        rangeB=3.0 (float): beginning of sampling range is x0 - ramgeB * std
        rangeE=3.0 (float): end of sampling range is x0 + ramgeE * std
    
    """
    
    def __init__(self, **kwargs):
        self._params = ["x0", "std", "nPoints", "rangeB", "rangeE"]
        self.x0 = 50e-9
        self.std = 10e-9
        self.nPoints = 30
        self.rangeB = 3.0
        self.rangeE = 3.0
        
        super(NormalDistribution, self).__init__(**kwargs)

    def __call__(self, x):
        """Returns distribution value at position x.
        
        Args:
            x (float): The point were to evaluate the distribution
        
        Returns:
            float: Distribution value at position x
            
        """
        res = 1.0 / (self.std * math.sqrt(2.0 * math.pi)) * \
            np.exp(- (x - self.x0) ** 2.0 / (2.0 * self.std ** 2.0))
        
        return res

    def GetDist(self):
        """Function returns distribution at nPoints sampling points.
        
        The sampling range is from max(1e-16, x0 - rangeB * std) to \
        x0 + rangeE * std where nPoints sampling points are taken.
        
        Returns:
            tuple_of_arrays: Distribution positions and values
            
        """
        xs = np.linspace(max(1e-12, self.x0 - self.rangeB * self.std), \
                         self.x0 + self.rangeE * self.std, self.nPoints)

        
        ws = self.__call__(xs)
        return xs, ws
            
class LogNormalDistribution(NormalDistribution):
    """This class presents lognormal distribution.
    
    Kwargs:
        x0=0.0 (float): mean
        std=1.0 (float): standard deviation
        nPoints=30 (int): number of points in distribution
        rangeB=0.0 (float): beginning of sampling range
        rangeE=10.0 (float): end of sampling range
    
    """
    def __init__(self, **kwargs):
        super(LogNormalDistribution, self).__init__(**kwargs)
   
    def __call__(self, x):
        """Returns distribution value at position x.
        
        Args:
            x (float): The point were to evaluate the distribution
        
        Returns:
            float: Distribution value at position x
            
        """
        #x02 = math.log(self.x0 ** 2.0 / math.sqrt(self.std + self.x0 ** 2.0))
        #std2 = math.sqrt(math.log(1.0 + self.std / self.x0 ** 2.0))
        x02 = np.log(self.x0)
        std2 = self.std
        
        #res = 1.0 / (std2 * math.sqrt(2.0 * math.pi) * x) * \
        #    np.exp(- (np.log(x) - x02) ** 2.0 / (2.0 * std2 ** 2.0))
        
        res = 1.0 / (std2 * math.sqrt(2.0 * math.pi)) * \
            np.exp(- (np.log(x) - x02) ** 2.0 / (2.0 * std2 ** 2.0))
        
       
        return res

    def GetDist(self):
        """Function returns distribution at nPoints sampling points.
        
        The sampling range is from max(1e-16, x0 - rangeB * std) to \
        x0 + rangeE * std where nPoints sampling points are taken.
        
        Returns:
            tuple_of_arrays: Distribution positions and values
            
        """
        xs = np.linspace(self.rangeB, self.rangeE, self.nPoints)

        ws = self.__call__(xs)
        return xs, ws

class LogNormalDistributionLocal(Core.ParamsBaseClass):
    """This class presents lognormal distribution with usual local parameters.
    
    Kwargs:
        xCoef (float): x-scale linear transformation parameter
        xOffset (float): x-offset value (new_x = x + xOffset)
        mu (float): location parameter
        sigma (float): scale parameter
        nPoint (int): number of points in distribution
        rangeB (float): beginning of sampling range
        rangeE (float): end of sampling range
        
    
    """
    def __init__(self, **kwargs):
        self._params = ["xCoef", "xOffset", "mu", "sigma", "nPoints", "distB", "distE"]
        self.xCoef = 1.0
        self.xOffset = 0.0
        self.mu = 0.0
        self.sigma = 1.0
        self.nPoints = 30
        self.distB = 0.0
        self.distE = 3.0
        
        super(LogNormalDistributionLocal, self).__init__(**kwargs)

    def __call__(self, x):
        """Returns distribution value at position x.
        
        Args:
            x (float): The point were to evaluate the distribution
        
        Returns:
            float: Distribution value at position x
            
        """
        
        x = self.xCoef * (x - self.xOffset)
        
        res = 1.0 / (self.sigma * math.sqrt(2.0 * math.pi) * x) * \
            np.exp(- (np.log(x) - self.mu) ** 2.0 / (2.0 * self.sigma ** 2.0))
        
        return res
    
    def GetDist(self):
        """Function returns distribution at nPoints sampling points.
        
        The sampling range is from max(1e-16, distB) to \
        distE where nPoints sampling points are taken.
        
        Returns:
            tuple_of_arrays: Distribution positions and values
            
        """
        xs = np.linspace(max(self.xOffset + 1e-16, self.distB), \
                         self.distE, self.nPoints)
        
        ws = self.__call__(xs)
        return xs, ws
        

if __name__ == "__main__":
    pass
#     import pylab as plt
#     dist = LogNormalDistribution(x0 = 10.0, std = 5.0)
#     print dist.GetParams()
#     xs, ws = dist.GetDist()
#     
#     plt.figure()
#     plt.plot(xs, ws, "x-")
#     plt.show()
#     
    


import numpy as np
import pytest
import LabPy
from NonlinearTMM import Material, SecondOrderNLTMM
    
@pytest.fixture(params = ["p", "s"])
def polP1(request):
    return request.param

@pytest.fixture(params = ["p", "s"])
def polP2(request):
    return request.param

@pytest.fixture(params = ["p", "s"])
def polGen(request):
    return request.param

@pytest.fixture(params = ["sfg", "dfg"])
def process(request):
    return request.param
    
@pytest.fixture()
def tmms(polP1, polP2, polGen, process):
    wlP1 = 700e-9
    wlP2 = 1500e-9
    I0P1 = 1.0
    I0P2 = 2.0
    dValues = {"d11": 1e-12, "d22": 2e-12, "d33": 3e-12, "d12": 4e-12, "d23": 5e-12}
    
    chi2 = LabPy.Chi2Tensor(**dValues)
    prism = LabPy.Material("Static", n = 1.5)
    metal = LabPy.Material("main/Ag/Johnson")
    crystal = LabPy.Material("special/gaussian_test")
    dielectric = LabPy.Material("Static", n = 1.0)
    crystalD = 1000e-6
    metalD = 5e-9
    
    tmmPy = LabPy.SecondOrderNLTmm(wlP1 = wlP1, wlP2 = wlP2, \
        polP1 = polP1, polP2 = polP2, polGen = polGen, \
        I0P1 = I0P1, I0P2 = I0P2, process = process)
    
    tmmPy.AddLayer(float("inf"), prism)
    tmmPy.AddLayer(metalD, metal)
    tmmPy.AddLayer(crystalD, crystal, chi2)
    tmmPy.AddLayer(float("inf"), dielectric)
    
    # C++
    prismCpp = Material.FromLabPy(prism)
    metalCpp = Material.FromLabPy(metal)
    crystalCpp = Material.FromLabPy(crystal)
    dielectricCpp = Material.FromLabPy(dielectric)
    crystalCpp.chi2.Update(**dValues)
    
    tmmCpp = SecondOrderNLTMM(process)
    tmmCpp.P1.SetParams(wl = wlP1, pol = polP1, I0 = I0P1)
    tmmCpp.P2.SetParams(wl = wlP2, pol = polP2, I0 = I0P2)
    tmmCpp.Gen.SetParams(pol = polGen)
    
    tmmCpp.AddLayer(float("inf"), prismCpp)
    tmmCpp.AddLayer(metalD, metalCpp)
    tmmCpp.AddLayer(crystalD, crystalCpp)
    tmmCpp.AddLayer(float("inf"), dielectricCpp)
    
    return tmmPy, tmmCpp
    
def test_PowerFlows(tmms):    
    tmmPy, tmmCpp = tmms
    betas = np.linspace(0.0, 0.99, 20)
    enhLayer, enhDist = 2, 20e-9
    srPy = tmmPy.Sweep(["betaP1", "betaP2"], [betas, betas], enhpos = (enhLayer, enhDist))
    srCpp = tmmCpp.Sweep("beta", betas, betas, outEnh = True, outAbs = True, layerNr = enhLayer, layerZ = enhDist)
    
    # P1
    np.testing.assert_allclose(srPy["iP1"], srCpp.P1.inc)
    np.testing.assert_allclose(srPy["rP1"], srCpp.P1.r)
    np.testing.assert_allclose(srPy["tP1"], srCpp.P1.t)
    np.testing.assert_allclose(srPy["IP1"].real, srCpp.P1.Ii)
    np.testing.assert_allclose(srPy["RP1"].real, srCpp.P1.Ir)
    np.testing.assert_allclose(srPy["TP1"].real, srCpp.P1.It)
    np.testing.assert_allclose(srPy["AP1"].real, srCpp.P1.Ia)
    np.testing.assert_allclose(srPy["enhP1"].real, srCpp.P1.enh)
    
    # P2
    np.testing.assert_allclose(srPy["iP2"], srCpp.P2.inc)
    np.testing.assert_allclose(srPy["rP2"], srCpp.P2.r)
    np.testing.assert_allclose(srPy["tP2"], srCpp.P2.t)
    np.testing.assert_allclose(srPy["IP2"].real, srCpp.P2.Ii)
    np.testing.assert_allclose(srPy["RP2"].real, srCpp.P2.Ir)
    np.testing.assert_allclose(srPy["TP2"].real, srCpp.P2.It)
    np.testing.assert_allclose(srPy["AP2"].real, srCpp.P2.Ia)
    np.testing.assert_allclose(srPy["enhP2"].real, srCpp.P2.enh)
    
    # Gen    
    np.testing.assert_allclose(srPy["iGen"], srCpp.Gen.inc)
    np.testing.assert_allclose(srPy["rGen"], srCpp.Gen.r)
    np.testing.assert_allclose(srPy["tGen"], srCpp.Gen.t)
    np.testing.assert_allclose(srPy["IGen"].real, srCpp.Gen.Ii)
    np.testing.assert_allclose(srPy["RGen"].real, srCpp.Gen.Ir)
    np.testing.assert_allclose(srPy["TGen"].real, srCpp.Gen.It)
    np.testing.assert_allclose(srPy["AGen"].real, srCpp.Gen.Ia)
    
    


"""This module contains low-level methods and classes for LabPy package.

"""

from bisect import bisect_left, bisect_right
import numpy as np
import scipy
import math
from time import time
from scipy.integrate import quad

#===============================================================================
# Methods
#===============================================================================

def Norm(vector):
    """Calculates vector 2nd order norm
    
    Args:
        vector (numpy.array): input x-array
    
    Returns:
        np.array: vector norm

    """
    if len(vector.shape) > 1:
        if vector.shape[1] != 3:
            raise Exception("Only vectors with length 3 supported.")
        return np.sqrt(abs(vector[:, 0]) ** 2.0 + abs(vector[:, 1]) ** 2.0 +  abs(vector[:, 2]) ** 2.0, dtype = complex).real
    else:        
        if len(vector) != 3:
            raise Exception("Only vectors with length 3 supported.")
        return np.sqrt(abs(vector[0]) ** 2.0 + abs(vector[1]) ** 2.0 + abs(vector[2]) ** 2.0, dtype = complex).real


def PickPoints(x, n, fr=float("-inf"), to=float("inf")):
    """Picks n points from array x, where fr <= x <= to
    
    Args:
        x (numpy.array): input x-array (must be sorted)
        n (int): number of points to pick
        fr (float, optional): requirement fr <= x
        to (float, optional): requirement x >= to  
    
    Returns:
        np.array: array of picked indices

    """
    
    indexFr = bisect_left(x, fr)
    indexTo = bisect_right(x, to)
    pointsTotal = indexTo - indexFr
    
    if pointsTotal < n:
        n = pointsTotal
        
    if n < 1:
        return np.array([])
    
    res = np.zeros(n, dtype = int)
    
    res[0] = indexFr
    pointsSel = 1

    pointsDensity = n / float(pointsTotal)

    for i in range(indexFr + 1, indexTo):
        shouldBeSelected = (i - indexFr + 1) * pointsDensity
        #print "cond", shouldBeSelected, pointsSel, int(shouldBeSelected + 1e-10)
        if pointsSel < int(shouldBeSelected + 1e-10):
            res[pointsSel] = i
            pointsSel += 1
            
    assert pointsSel == n
    return res
    
    
def complex_quadrature(func, a, b, **kwargs):
    """Integrates complex function func from a to b
    
    Args:
        func (function): complex function to integrate
        a (float): integration parameter
        b (float): integration parameter
        kwargs (dict, optional): optimal parameters for quad  
    
    Returns:
        tuple: (complex integration value, real integration value, imaginary integral value)

    """
    def real_func(x):
        return scipy.real(func(x))
    def imag_func(x):
        return scipy.imag(func(x))
    
    real_integral = quad(real_func, a, b, **kwargs)
    imag_integral = quad(imag_func, a, b, **kwargs)
    return (real_integral[0] + 1j*imag_integral[0], real_integral[1:], imag_integral[1:])


def FindZeros(func, dfunc, a, b, zerosMaxDepth = 5, disp = True, **kwargs):
    """Finds all zeros of function (in range a to b) by argument principle method
    
    Args:
        argPrintcipleFunc (function): function of argument principle
        a (complex): integration parameter
        b (complex): integration parameter
        zerosMaxDepth (int): maximum recursion depth
        disp (bool): display debug data
        kwargs (dict, optional): optimal parameters for integration  
    
    Returns:
        list of complex: the list of zeros

    """
    def _ZerosInBox(ax, bx):
        start = time()
        
        a1x, a2x = min(ax.real, bx.real), max(ax.real, bx.real)
        b1x, b2x = min(ax.imag, bx.imag), max(ax.imag, bx.imag)
        
        func1 = lambda x: argPrintcipleFunc(a1x + 1.0j * x)
        func2 = lambda x: argPrintcipleFunc(x + 1.0j * b2x)
        func3 = lambda x: argPrintcipleFunc(a2x + 1.0j * x)
        func4 = lambda x: argPrintcipleFunc(x + 1.0j * b1x)
        
        integArgs = kwargs
        integral = 1.0j * complex_quadrature(func1, b1x, b2x, **integArgs)[0]
        integral += complex_quadrature(func2, a1x, a2x, **integArgs)[0]
        integral += 1.0j * complex_quadrature(func3, b2x, b1x, **integArgs)[0]
        integral += complex_quadrature(func4, a2x, a1x, **integArgs)[0]
        
        res = -integral / (2.0j * math.pi)
        
        if disp:
            print("zeros in box", ax, bx, res)
            print("time for box:", time() - start)
        
        
        
        return res
    
    
    def _DoSearch(a, b, depth = 0):
        zerosRaw = _ZerosInBox(a, b).real
        print("DoSearch", a, b, depth, "zeros", zerosRaw)
        zeros = int(round(zerosRaw))
        if zeros <= 0:
            return []
        
        if depth > zerosMaxDepth:
            return [(a + b) / 2.0]
        
        a1, a3 = min(a.real, b.real), max(a.real, b.real)
        b1, b3 = min(a.imag, b.imag), max(a.imag, b.imag)
        a2, b2 = (a1 + a3) / 2.0, (b1 + b3) / 2.0
     
        res = []
        res += _DoSearch(complex(a1, b1), complex(a2, b2), depth + 1)
        res += _DoSearch(complex(a2, b1), complex(a3, b2), depth + 1)
        res += _DoSearch(complex(a1, b2), complex(a2, b3), depth + 1)
        res += _DoSearch(complex(a2, b2), complex(a3, b3), depth + 1)
        return res
        
    argPrintcipleFunc = lambda x: dfunc(x) / func(x)
    res = _DoSearch(a, b, 0)
    return res
    
def ExtraInterpolate(xs, ys, x):
    if len(xs) == 1:
        return ys[0]
    elif len(xs) == 2:
        return (x * ys[0] - x * ys[1] + xs[0] * ys[1] - xs[1] * ys[0]) / (xs[0] - xs[1])
    elif len(xs) == 3:
        return ((ys[2] * xs[1] - ys[1] * xs[2] + x * (ys[1] - ys[2])) * xs[0] ** 2 + (-ys[2] * xs[1] ** 2 + ys[1] * xs[2] ** 2 - x ** 2 * (ys[1] - ys[2])) * xs[0] + (ys[0] * xs[2] - x * (ys[0] - ys[2])) * xs[1] ** 2 + (-ys[0] * xs[2] ** 2 + x ** 2 * (ys[0] - ys[2])) * xs[1] - xs[2] * x * (ys[0] - ys[1]) * (x - xs[2])) / (xs[1] - xs[2]) / (xs[0] - xs[2]) / (xs[0] - xs[1])
    else:
        raise NotImplementedError("len = %d" % (len(xs)))

def RotationMatrixX(phi):
    """Rotation matrix around X-axis
    
    Args:
        phi (float): roatation angle
        
    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[1.0, 0.0, 0.0],
                    [0.0, np.cos(phi), -np.sin(phi)],
                    [0.0, np.sin(phi), np.cos(phi)]], dtype = float)
    return res

def RotationMatrixY(phi):
    """Rotation matrix around Y-axis
    
    Args:
        phi (float): roatation angle
        
    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[np.cos(phi), 0.0, np.sin(phi)],
                    [0.0, 1.0, 0.0],
                    [-np.sin(phi), 0.0, np.cos(phi)]], dtype = float)
    return res

def RotationMatrixZ(phi):
    """Rotation matrix around Z-axis
    
    Args:
        phi (float): roatation angle
        
    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[np.cos(phi), -np.sin(phi), 0.0],
                    [np.sin(phi), np.cos(phi), 0.0],
                    [0.0, 0.0, 1.0]], dtype = float)
    return res

def CartesianProduct(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:,0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in range(1, arrays[0].size):
            out[j*m:(j+1)*m,1:] = out[0:m,1:]
    return out

#===============================================================================
# Classes
#===============================================================================

class ParamsBaseClass(object):
    """Inheritance of this class will provide SetParams and GetParams methods.
    
    Args:
        kwargs: optional arguments, currently unused
    
    """
    
    def __init__(self, params = None, **kwargs):
        if params is not None:
            self._params = params  
            
        if not hasattr(self, "friendlyNames"):
            self.friendlyNames = self._params[:]
             
        self.SetParams(**kwargs)

        
    def SetParams(self, **kwargs):
        """Sets parameters from kwargs dictionary.
        
        Args:
            kwargs (dict): dictionary of parameters
            
        Raises:
            ValueError: if parameter name is not in _params list
        
        """
        
        for k, v in kwargs.items():
            if not k in self._params:
                raise ValueError("Unknown kwargs: %s" % str(k))
            setattr(self, k, v)
            
        
    def GetParams(self, asList = False):
        """Returns values of parameters in _params.

        Args:
            asList (bool, optional = False): Whether return params as list or as dict. 

        Returns:
            dixt: dictionary of parameters and values
        
        """
        if asList:
            res = []
            for param, friendlyName in zip(self._params, self.friendlyNames):
                res.append((param, friendlyName, getattr(self, param)))
                
        else:
            res = {}
            for param in self._params:
                res[param] = getattr(self, param)
        return res   
    
if __name__ == "__main__":
    
    for c1 in np.arange(1, 1000, 1):
        a = np.linspace(0.0, 100.0, c1)
        for c2 in np.arange(1, c1, 1):
            newa = PickPoints(a, c2)
            
    #newa = PickPoints(range(39), 31)

            
    